#! /usr/bin/env python
class DisjointSet:
    '''``DisjointSet`` class, implemented using the Up-Tree data structure for amortized O(1) find and union operations'''
    def __init__(self, initial=None):
        '''`DisjointSet` constructor
        
        Args:
            ``initial`` (iterable): Elements with which to initialize the ``DisjointSet`` (each element will be in its own set)
        '''
        self.parent = dict() # parent[u] = parent of node u
        self.num_below = dict() # num_below[u] = number of nodes below u (including u) (only current for sentinels)
        if initial is not None:
            for x in initial:
                self.add(x)

    def __contains__(self, x):
        '''Check if an element ``x`` exists in this ``DisjointSet``

        Args:
            ``x``: The element to check

        Returns:
            ``bool``: ``True`` if ``x`` exists in this ``DisjointSet``, otherwise ``False``
        '''
        try:
            return x in self.parent
        except:
            return False

    def __len__(self):
        '''Return the number of elements in this ``DisjointSet``

        Returns:
            ``int``: The number of elements contained within this ``DisjointSet``
        '''
        return len(self.parent)

    def add(self, x):
        '''Add a new element ``x`` to this ``DisjointSet`` as a sentinel node

        Args:
            ``x``: The element to insert
        '''
        if x in self:
            raise ValueError("Node already exists: %s"%x)
        self.parent[x] = None; self.num_below[x] = 1

    def remove(self, x):
        '''Remove the element ``x`` from this ``DisjointSet``

        Args:
            ``x``: The element to remove
        '''
        if x not in self:
            raise ValueError("Node not found: %s"%x)
        p = self.parent[x]
        if p is not None:
            p = self.parent[x]; self.num_below[p] -= 1
        for e in self.parent:
            if self.parent[e] == x:
                self.parent[e] = p
        del self.parent[x]; del self.num_below[x]

    def find(self, x):
        '''Return the sentinel node of the element ``x``. Implements path compression along the search

        Args:
            ``x``: The element to find

        Returns:
            The sentinel node of ``x``
        '''
        if x not in self:
            raise ValueError("Node not found: %s"%x)
        explored = Queue(); curr = x
        while self.parent[curr] is not None:
            explored.put(curr); curr = self.parent[curr]
        while not explored.empty():
            self.parent[explored.get()] = curr
        return curr

    def union(self, x, y):
        '''Union the sets containing ``x`` and ``y``. Implements Union-By-Size

        Args:
            ``x``: One of the two elements whose sets will be unioned
            ``y``: One of the two elements whose sets will be unioned
        '''
        if x not in self:
            raise ValueError("Node not found: %s"%x)
        if y not in self:
            raise ValueError("Node not found: %s"%y)
        sx = self.find(x); sy = self.find(y)
        if sx == sy:
            return
        if self.num_below[sx] > self.num_below[sy]:
            self.parent[sy] = sx; self.num_below[sx] += (self.num_below[sy] + 1)
        else:
            self.parent[sx] = sy; self.num_below[sy] += (self.num_below[sx] + 1)



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
setup(
    name='niemads',  # Required
    version='1.0.1',  # Required
    description='NiemaDS: Non-standard data structures for Python 2 and 3',  # Required
    long_description='NiemaDS is a Python library that implements useful non-standard data structures.',  # Optional
    long_description_content_type='text/plain',  # Optional (see note above)
    url='https://github.com/niemasd/TreeDS',  # Optional
    author='Niema Moshiri',  # Optional
    author_email='niemamoshiri@gmail.com',  # Optional
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='data structures fast',  # Optional
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),  # Required
    extras_require={  # Optional
        'dev': ['check-manifest'],
    },
    project_urls={  # Optional
        'Bug Reports': 'https://github.com/niemasd/NiemaDS/issues',
        'Source': 'https://github.com/niemasd/NiemaDS',
    },
)
